#include <cctype>
#include <cstdarg>
#include <ctime>
#include <errno.h>
#include <iostream>
#include <unistd.h>
#include <sys/stat.h>
#include <poll.h>

using std::cout;
using std::cerr;
using std::endl;

#define TELOPTS
#define TELCMDS
#include <arpa/telnet.h>

#include "playertask.hh"
#include "playerdata.hh"

////////////

playertask::playertask(int s) : sock(s), got_iac(0), flowmode(f_normal)
{
	data = new playerdata(*this);
	
	telnetsend(
		IAC, DO, TELOPT_TTYPE,
		IAC, DO, TELOPT_NAWS,
		IAC, DO, TELOPT_LFLOW,
		-1);
		
	setlflow();
	
	cout << "[" << sock << "]\n";
	
	data->startup();
}

void playertask::setlflow()
{
	hideprompt();
	
	flowmode = f_normal;
	
	telnetsend(
		IAC, WONT, TELOPT_SGA,
		IAC, WONT, TELOPT_ECHO,
		-1);
}

void playertask::setprompt(const string &prompt, unsigned maxlength, bool visible, const string &charset)
{
	inputcharset = charset;
	inputprompt  = prompt;
	inputmaxlen  = maxlength;
	flowmode     = visible ? f_visible : f_hidden;
	
	telnetsend(
		IAC, WILL, TELOPT_SGA,
		IAC, WILL, TELOPT_ECHO,
		-1);
		
	inputhidden = true;
	
	showprompt();
}

void playertask::hideprompt() const
{
	if(!inputhidden)
	{
		inputhidden = true;
		
		data->colour(-1);
		
		private_echo("\33[0;37;40m\r\33[K");
	}
}

void playertask::showprompt() const
{
	if(inputhidden)
	{
		data->colour(-1);
		
		string Buf = input;
		
		if(Buf.size() > inputmaxlen)
			Buf.erase(inputmaxlen);
		
		string Bufs(inputmaxlen - Buf.size(), ' ');
		string Bufb(Bufs.size(),              '\b');
		
		private_echo("\r\33[0;37;40m" + inputprompt + "\33[0;1;44m" + Buf + Bufs + Bufb);
		
		inputok();
		
		inputhidden = false;
	}
}

playertask::~playertask()
{
	cout << "[~" << sock << "]\n";
	
	delete data;
	
	shut();
}

void playertask::private_echo(const string &s) const
{
	//SendQueue.Feed((const unsigned char *)s.data(), s.size());
}

void playertask::echo(const string &s, bool endcr) const
{
	string tmp = s;
	if(endcr) tmp += "\n";
	
	if(tmp.size())
	{
		if(flowmode != f_normal)
			hideprompt();
	
		private_echo(tmp);
		
		if(flowmode != f_normal && tmp[tmp.size()-1] == '\n')
			showprompt();
	}
}

bool playertask::tick()
{
	return (sock >= 0);
}

unsigned playertask::numpolls() const
{
    return 1;
}

void playertask::writepolls(struct pollfd *arr)
{
    arr[0].fd = sock;
    arr[0].events = POLLIN;
    if(SendQueue.GetSize())
    	arr[0].events |= POLLOUT;
}

void playertask::checkpolls(const struct pollfd *arr)
{
    if(arr[0].revents & POLLERR)
    {
   		extern const char *const sys_errlist[];
   		cerr << "[!" << sock << "]: " << sys_errlist[errno] << endl;
    }
    if(arr[0].revents & POLLHUP)
    {
    	close(sock);
    	sock = -1;
    }
    if(arr[0].revents & POLLOUT)
    {
    	const unsigned char *p = SendQueue.GetBegin();
    	
    	int n = write(sock, p, SendQueue.GetSize());
    	if(n > 0)
    		SendQueue.Eat(n);
    }
    if(arr[0].revents & POLLIN)
    {
    	unsigned char Buf[2048];
    	
    	int n = read(sock, Buf, sizeof(Buf));
    	for(int a=0; a<n; ++a)
    	{
    		unsigned char ch = Buf[a];
    		
    		switch(got_iac)
    		{
    			case 0:
    			{
    				if(ch == IAC)
    					++got_iac;
    				else
    				{
                		switch(ch)
                		{
                			case '':
                			case '':
                				if(input.size() > 0)
                				{
                					input = input.substr(0, input.size()-1);
                					if(flowmode != f_normal)
                						private_echo("\b \b");
                				}
                				break;
                			case '\r':
                			case '\n':
                			{
                				if(flowmode != f_normal)
                				{
                					private_echo("\33[0;37;40m\r\n");
                					data->colour(-1);
                					setlflow();
                				}
                				
    							cout << "[" << sock << "]: '" << input << '\\' << (ch=='\r' ? 'r' : 'n') << "'\n";            				
    							
                			    data->handleinput();
                				
                				inputok();
                				break;
                			}
                			default:
                				if(ch >= ' ')
                    			{
                   				//	cout << "Gotch: '" << ch << "'\n";
                    				if(flowmode == f_normal)
                    				{
        	            				if(ch >= ' ')input += ch;
        	            			}
        	            			else
        	            			{
        	            				if(input.size() < inputmaxlen
        	            				&& inputcharset.find(ch) != inputcharset.npos)
        	            				{
        	            					input += ch;
        	            					private_echo(string("") + char(flowmode == f_hidden ? '*' : ch));
        	            				}
        	            			}
                    			}
                		}
                	}
    				break;
    			}
    			case 1: // got_iac == 1
    				switch(ch)
    				{
    					case DO:
    					case DONT:
    					case WILL:
    					case WONT:
    					case SB:
    						got_iac = ch;
    						break;
    					case IP:
    						input = "";
    						// fall-through
    						echo("^C");
    					case AYT:
    						// fall-through
    					default:
    						cout << "Got: IAC " << telcmds[ch-xEOF] << endl;
    						got_iac = 0;
    						break;
    				}
    				break;
    			case SB:
    				got_iac = DM;
    				got_sb = ch;
    				got_si = 0;
    				break;
    			case DM:
    				switch(got_sb)
    				{
    					case TELOPT_NAWS:
    						got_data[got_si++] = ch;
    						if(got_si == 4)
    						{
    							unsigned x = got_data[1] + (got_data[0]<<8);
    							unsigned y = got_data[3] + (got_data[2]<<8);
    							got_iac = 0;
    							data->resize(x, y);
    							cout << "[<" << sock << ">]: " << x << "x" << y << endl;
    					    }
    						break;
    					default:
    						got_iac = 0;
    				}
    				break;
    			default:
    				switch(ch)
    				{
    					default:
    					//	if(got_iac == DO)  telnetsend(IAC, WILL, ch, -1);
    					//	if(got_iac == DONT)telnetsend(IAC, WONT, ch, -1);
    						cout << "Got: IAC " << telcmds[got_iac-xEOF] << " " << telopts[ch] << endl;
    				}
    				got_iac = 0;
    		}
    	}
    }
}

void playertask::operator += (const event &ev)
{
	data->takecareof(ev);
}

void playertask::shut()
{
	usleep(100000);
	close(sock);
	sock = -1;
}

void playertask::inputok() const
{
	telnetsend(IAC, GA, -1);
	if(flowmode == f_normal)
		data->colour(7);
}

void playertask::telnetsend(int dummy, ...) const
{
	va_list ap;
	va_start(ap, dummy);
	
	int a;
	string Buf;
	
	Buf += (char)dummy;
	
	while((a = va_arg(ap, int)) >= 0)
		Buf += (char)a;
		
	private_echo(Buf);
}
